#ifndef SEAFILE_CLIENT_PAINT_UTILS_H_
#define SEAFILE_CLIENT_PAINT_UTILS_H_

#include <QFont>
#include "utils-mac.h"
#include <QIcon>

QString fitTextToWidth(const QString& text, const QFont& font, int width);

QFont zoomFont(const QFont& font_in, double ratio);

QFont changeFontSize(const QFont& font_in, int size);

int textWidthInFont(const QString text, const QFont& font);

int textHeightInFont(const QString text, const QFont& font);

double globalDevicePixelRatio();

#endif // SEAFILE_CLIENT_PAINT_UTILS_H_
